@ assembly "optimized" blitter and copy functions
@ all pointers must be word-aligned

@ (c) Copyright 2006, notaz
@ All Rights Reserved


@ Convert 0000bbb0 ggg0rrr0
@ to      0000rrr0 ggg0bbb0

@ r2,r3 - scratch, lr = 0x000F000F
.macro convRGB444 reg
    and     r2,   \reg, lr         @ r2=red
    and     r3,   \reg, lr, lsl #8 @ r3=blue
    and     \reg, \reg, lr, lsl #4 @ green stays in place
    orr     \reg, \reg, r2, lsl #8 @ add red back
    orr     \reg, \reg, r3, lsr #8 @ add blue back
.endm

.global vidConvCpyRGB444 @ void *to, void *from, int pixels

vidConvCpyRGB444:
    stmfd   sp!, {r4-r11,lr}

    mov     r12, r2, lsr #4 @ repeats
    mov     lr, #0xF0000
    orr     lr, lr, #0xF    @ lr == pattern 0x000F000F


.loopRGB444:
	subs    r12, r12, #1

    @ I first thought storing multiple registers would be faster,
    @ but this doesn't seem to be the case, probably because of
    @ slow video memory we are dealing with
 	ldmia	r1!, {r4-r11}
    convRGB444 r4
    str     r4, [r0], #4
    convRGB444 r5
    str     r5, [r0], #4
    convRGB444 r6
    str     r6, [r0], #4
    convRGB444 r7
    str     r7, [r0], #4
    convRGB444 r8
    str     r8, [r0], #4
    convRGB444 r9
    str     r9, [r0], #4
    convRGB444 r10
    str     r10, [r0], #4
    convRGB444 r11
    str     r11, [r0], #4

    bgt     .loopRGB444


    ldmfd   sp!, {r4-r11,lr}
    bx      lr


@ Convert 0000bbb0 ggg0rrr0
@ to      rrr00ggg 000bbb00

@ r2,r3 - scratch, lr = 0x07800780
.macro convRGB565 reg
    and     r2,   \reg, lr,  lsr #7  @ r2=red
    and     r3,   \reg, lr,  lsl #1  @ r3=blue
    and     \reg, lr,   \reg,lsl #3  @ green stays, but needs shifting
    orr     \reg, \reg, r2,  lsl #12 @ add red back
    orr     \reg, \reg, r3,  lsr #7  @ add blue back
.endm

.global vidConvCpyRGB565 @ void *to, void *from, int pixels

vidConvCpyRGB565:
    stmfd   sp!, {r4-r11,lr}

    mov     r12, r2, lsr #4 @ repeats
    mov     lr, #0x07800000
    orr     lr, lr, #0x780  @ lr == pattern 0x07800780

.loopRGB565:
	subs    r12, r12, #1

 	ldmia	r1!, {r4-r11}
    convRGB565 r4
    str     r4, [r0], #4
    convRGB565 r5
    str     r5, [r0], #4
    convRGB565 r6
    str     r6, [r0], #4
    convRGB565 r7
    str     r7, [r0], #4
    convRGB565 r8
    str     r8, [r0], #4
    convRGB565 r9
    str     r9, [r0], #4
    convRGB565 r10
    str     r10, [r0], #4
    convRGB565 r11
    str     r11, [r0], #4

    bgt     .loopRGB565

    ldmfd   sp!, {r4-r11,lr}
    bx      lr


@ Convert 0000bbb0 ggg0rrr0 0000bbb0 ggg0rrr0
@ to      00000000 rrr00000 ggg00000 bbb00000 ...

@ r2,r3 - scratch, lr = 0x0000F000
@ rin - src reg, rout - dest reg (can be same for both; rout can be r3)
.macro convRGB32_l rout rin
    and     r2,    \rin,  lr,   lsr #12 @ r2=red
    and     r3,    \rin,  lr,   lsr #4  @ r3=blue
    orr     r2,    r3,    r2,   lsl #24
    and     \rout, lr,    \rin, lsl #8  @ green stays, but needs shifting
    orr     \rout, \rout, r2,   lsr #4  @ add red+blue back
.endm

@ r2,r3 - scratch, lr = 0x0000F000
@ rin - src reg, rout - dest reg (can be same for both)
.macro convRGB32_h rout rin
    and     r2,    \rin,  lr,   lsl #4  @ r2=red
    mov     r3,    \rin,        lsr #24 @ r3=blue
    and     \rout, lr,    \rin, lsr #8  @ green
    orr     \rout, \rout, r2,   lsl #4
    orr     \rout, \rout, r3,   lsl #4
.endm

.global vidConvCpyRGB32 @ void *to, void *from, int pixels

vidConvCpyRGB32:
    stmfd   sp!, {r4-r11,lr}

    mov     r12, r2, lsr #4 @ repeats
    mov     lr, #0x0000F000

.loopRGB32:
	subs    r12, r12, #1

 	ldmia	r1!, {r4-r11}
    convRGB32_l r3 r4
    str     r3, [r0], #4
    convRGB32_h r4 r4
    str     r4, [r0], #4

    convRGB32_l r3 r5
    str     r3, [r0], #4
    convRGB32_h r5 r5
    str     r5, [r0], #4

    convRGB32_l r3 r6
    str     r3, [r0], #4
    convRGB32_h r6 r6
    str     r6, [r0], #4

    convRGB32_l r3 r7
    str     r3, [r0], #4
    convRGB32_h r7 r7
    str     r7, [r0], #4

    convRGB32_l r3 r8
    str     r3, [r0], #4
    convRGB32_h r8 r8
    str     r8, [r0], #4

    convRGB32_l r3 r9
    str     r3, [r0], #4
    convRGB32_h r9 r9
    str     r9, [r0], #4

    convRGB32_l r3 r10
    str     r3, [r0], #4
    convRGB32_h r10 r10
    str     r10,[r0], #4

    convRGB32_l r3 r11
    str     r3, [r0], #4
    convRGB32_h r11 r11
    str     r11,[r0], #4

    bgt     .loopRGB32

    ldmfd   sp!, {r4-r11,lr}
    bx      lr


