// some debug stuff:
// custom ExceptionHandler with crash address guessing code
// and DEBUGPRINT macros

// (c) Copyright 2006, notaz
// All Rights Reserved


#include <e32svr.h> // RDebug


static TPtrC exception_names[] = {
	_L("General"),
	_L("IntegerDivideByZero"),
	_L("SingleStep"),
	_L("BreakPoint"),
	_L("IntegerOverflow"),
	_L("BoundsCheck"),
	_L("InvalidOpCode"),
	_L("DoubleFault"),
	_L("StackFault"),
	_L("AccessViolation"),
	_L("PrivInstruction"),
	_L("Alignment"),
	_L("PageFault"),
	_L("FloatDenormal"),
	_L("FloatDivideByZero"),
	_L("FloatInexactResult"),
	_L("FloatInvalidOperation"),
	_L("FloatOverflow"),
	_L("FloatStackCheck"),
	_L("FloatUnderflow"),
	_L("Abort"),
	_L("Kill"),
	_L("DataAbort"),
	_L("CodeAbort"),
	_L("MaxNumber"),
	_L("InvalidVector"),
	_L("UserInterrupt"),
	_L("Unknown")
};


static void getASpace(TUint *code_start, TUint *code_end, TUint *stack_start, TUint *stack_end)
{
	TUint pc, sp;
	RChunk chunk;
	TFullName chunkname;
	TFindChunk findChunk(_L("*"));

	asm volatile ("str pc, %0" : "=m" (pc) );
	asm volatile ("str sp, %0" : "=m" (sp) );

	while( findChunk.Next(chunkname) != KErrNotFound ) {
		chunk.Open(findChunk);
		if((TUint)chunk.Base()+chunk.Bottom() < pc && pc < (TUint)chunk.Base()+chunk.Top()) {
			if(code_start) *code_start = (TUint)chunk.Base()+chunk.Bottom();
			if(code_end)   *code_end   = (TUint)chunk.Base()+chunk.Top();
		} else
		if((TUint)chunk.Base()+chunk.Bottom() < sp && sp < (TUint)chunk.Base()+chunk.Top()) {
			if(stack_start) *stack_start = (TUint)chunk.Base()+chunk.Bottom();
			if(stack_end)   *stack_end   = (TUint)chunk.Base()+chunk.Top();
		}
		chunk.Close();
	}
}

// our very own exception handler
void ExceptionHandler(TExcType exc)
{
	TUint lr, sp, i;
	TUint stack_end	= 0;				// ending address of our stack chunk
	TUint code_start = 0, code_end = 0; // starting and ending addresses of our code chunk
	TUint guessed_address = 0;

	asm volatile ("str lr, %0" : "=m" (lr) );
	asm volatile ("str sp, %0" : "=m" (sp) );

	// first get some info about the chunks we live in
	getASpace(&code_start, &code_end, 0, &stack_end);

	// now we begin some black magic tricks
	// we go up our stack until we pass our caller address
	for(; sp < stack_end; sp += 4)
		if(*(TUint *)sp == lr) break;

	// there might be mirored caller address
	for(i = sp + 4; i < sp + 0x500 && i < stack_end; i += 4)
		if(*(TUint *)i == lr) { sp = i; break; }

	// now we continue until we pass 2 system addresses
	for(i = 0; sp < stack_end; sp += 4)
		if((*(TUint *)sp >> 28) == 5 && ++i == 2) break;

	if(i == 2) {
		// there should be 2 addresses with 0x14 bytes between then. The second one should be the offending address.
		for(sp += 0x18; sp < stack_end; sp += 4) {
			if(code_start < *(TUint *)sp && *(TUint *)sp < code_end && code_start < *(TUint *)(sp-0x14) && *(TUint *)(sp-0x14) < code_end)
				{ guessed_address = *(TUint *)sp; break; }
		}
	}

	// output the info
	TUint exec_show = exc;
	if(exec_show > 27) exec_show = 27;
	RDebug::Print(_L("!!!Exception %i (%S) @ 0x%08x (guessed; relative=0x%08x)"), exc, &exception_names[exec_show], guessed_address, guessed_address - code_start);

	//RDebug::Print(_L("Stack dump:"));
	//asm volatile ("str sp, %0" : "=m" (sp) );
	//for(TUint i = sp+0x400; i >= sp-16; i-=4)
	//	RDebug::Print(_L("%08x: %08x"), i, *(int *)i);

	// more descriptive replacement of "KERN-EXEC 3" panic
	TBuf<48> error;
	error.Format(_L("K-EX3: %S"), &exception_names[exec_show]);
	User::Panic(error, exc);
}


// c string dumper for RDebug::Print()
#ifdef __DEBUG_PRINT
	#define DEBUGPRINT(x...) RDebug::Print(x)
	static	TBuf<512> sTextBuffer;
	TDesC* DO_CONV(const char* s)
	{
		TPtrC8	text8((TUint8*) (s));
		sTextBuffer.Copy(text8);
		return &sTextBuffer;
	}

	#include <stdarg.h> // va_*
	#include <stdio.h>  // vsprintf

	// debug print from c code
	extern "C" void dprintf(char *format, ...)
	{
		va_list args;
		char    buffer[512];

		va_start(args,format);
		vsprintf(buffer,format,args);
		va_end(args);

		RDebug::Print(_L("%S"), DO_CONV(buffer));
	}
#else
	#define DEBUGPRINT(x...)
#endif

